# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
CLI for interacting with individual pybuild files
"""

import argparse
import json
import logging
import lzma
import os
import pickle
import re
import sys
from contextlib import redirect_stdout
from io import StringIO
from types import SimpleNamespace
from typing import cast

from .cache import create_cache_str, pybuild_dumper
from .execute import execute
from .globals import env
from .log import init_logger
from .package import PhaseState
from .pybuild import FullPybuild
from .repo import get_repo, get_repo_root
from .repo.loader import __load_file, __load_installed


def wrapper_pybuild(args):
    if args.env:
        env.deserialize(bytes.fromhex(args.env))

    # Set info variables
    from pybuild.info import _set_info

    _set_info(args.pybuild_file)

    def path_in_installed_db(path: str) -> bool:
        try:
            return bool(
                os.path.commonpath(
                    [os.path.abspath(os.path.normpath(path)), env.prefix().INSTALLED_DB]
                )
                == env.prefix().INSTALLED_DB
            )
        except ValueError:
            return False

    # Redirect stdout when loading and raise an exception complaining
    # about output in the global/init scope
    output = StringIO()
    with redirect_stdout(output):
        pkg: FullPybuild
        if env.PREFIX_NAME and path_in_installed_db(args.pybuild_file):
            with open(
                os.path.join(os.path.dirname(args.pybuild_file), "REPO"), "r"
            ) as file:
                sys.path.append(get_repo(file.read().strip()).location)
            pkg = __load_installed(args.pybuild_file)
        else:
            repo_root = cast(str, get_repo_root(args.pybuild_file))
            sys.path.append(repo_root)
            pkg = __load_file(args.pybuild_file)
    if output.getvalue():
        raise Exception(
            "Pybuilds should not produce output in the global scope or __init__!"
        )

    if args.state_dir and os.path.exists(args.state_dir):
        state_file = os.path.join(args.state_dir, "state.pickle")
        if os.path.exists(state_file):
            with open(state_file, "rb") as b_file:
                pkg.__dict__ = pickle.load(b_file)

    if args.initial_state:
        state = PhaseState.from_json(json.loads(args.initial_state))
        pkg.__dict__.update(state.__dict__)

    def pkg_func(pkg, name):
        pkg.execute = execute
        func = getattr(pkg, name)
        return func()

    for command in args.command:
        if command == "cache":
            print(create_cache_str(pkg, cache=True))
        elif command == "unpack":
            pkg_func(pkg, "src_unpack")
        elif command == "prepare":
            pkg_func(pkg, "src_prepare")
        elif command == "install":
            pkg_func(pkg, "src_install")
        elif command == "postinst":
            pkg_func(pkg, "pkg_postinst")
        elif command == "pretend":
            pkg_func(pkg, "pkg_pretend")
        elif command == "nofetch":
            pkg_func(pkg, "pkg_nofetch")
        elif command == "can-update-live":
            result = pkg_func(pkg, "can_update_live")
            if not isinstance(result, bool):
                raise RuntimeError(
                    f"can_update_live returned unexpected result {result}"
                )
            sys.exit(int(result) * 142)
        elif command == "validate":
            pkg._validate()

    if args.state_dir:
        with open(os.path.join(args.state_dir, "state.pickle"), "wb") as b_file:
            pickle.dump(pkg.__dict__, b_file)

        with open(os.path.join(args.state_dir, "environment.xz"), "wb") as b_file:
            # Keys are sorted to produce consistent results and
            # easy to read commits in the DB
            dictionary = pkg.__class__.__dict__.copy()
            dictionary.update(pkg.__dict__)
            dictionary = dict(
                filter(
                    lambda elem: not elem[0].startswith("_") and elem[0] != "execute",
                    dictionary.items(),
                )
            )
            jsonstr = json.dumps(dictionary, default=pybuild_dumper, sort_keys=True)
            b_file.write(lzma.compress(str.encode(jsonstr)))


def wrapper_module(args):
    if args.env:
        env.deserialize(bytes.fromhex(args.env))

    if args.initial_state:
        state = PhaseState.from_json(json.loads(args.initial_state))

    name, _ = os.path.splitext(os.path.basename(args.module_file))
    module = os.path.basename(os.path.dirname(args.module_file))
    sys.path.append(os.path.dirname(os.path.dirname(args.module_file)))
    with open(args.module_file, "r", encoding="utf-8") as file:
        tmp_globals = {
            "__builtins__": __builtins__,
            "__name__": name,
            "__package__": module,
        }
        code = compile(file.read(), args.module_file, "exec")
        exec(code, tmp_globals)

    do_functions = {}
    descriptions = {}
    describe_options = {}
    describe_parameters = {}
    for globname in tmp_globals:
        if globname.startswith("do_"):
            name = re.sub("^do_", "", globname)
            do_functions[name] = tmp_globals[globname]
            descriptions[name] = tmp_globals[globname].__doc__
        match = re.match("^describe_(.*)_options$", globname)
        if match:
            describe_options[match.group(1)] = tmp_globals[globname]()
        match = re.match("^describe_(.*)_parameters$", globname)
        if match:
            describe_parameters[match.group(1)] = tmp_globals[globname]()

    if args.command == "load":
        name = os.path.basename(args.module_file)
        name, _ = os.path.splitext(name)
        dictionary = {
            "name": name,
            "desc": tmp_globals.get("__doc__"),
            "functions": {
                name: {
                    "desc": descriptions[name],
                    "options": describe_options.get(name),
                    "parameters": describe_parameters.get(name),
                }
                for name in descriptions
            },
        }
        print(json.dumps(dictionary))
    elif args.command == "execute" and args.module_func:
        if args.module_func == "update":
            do_functions[args.module_func](state)
        else:
            function_args = None
            if args.args:
                function_args = SimpleNamespace(**json.loads(args.args))
            do_functions[args.module_func](state, function_args)


def main():
    """
    Wrapper script for directly loading pybuild and module files.
    This should always be invoked using the executable sandbox,
    and is not intended to be invoked manually.
    """
    parser = argparse.ArgumentParser(description=main.__doc__)
    parser.add_argument("--debug", help="Enables debug traces", action="store_true")
    parser.add_argument("--verbosity", help="verbosity level")
    parser.add_argument(
        "--env",
        help="Json-encoded dictionary of values to apply to the global environment",
    )
    subparsers = parser.add_subparsers()
    pybuild = subparsers.add_parser("pybuild", help="interact with pybuild files")
    module = subparsers.add_parser("module", help="interact with module files")
    pybuild.add_argument("pybuild_file", metavar="<pybuild file>")
    pybuild.add_argument("command", metavar="<command>", nargs="+")
    pybuild.add_argument(
        "--state-dir",
        help="The path of a directory to be used to store state information",
    )
    pybuild.add_argument(
        "--initial-state",
        help="A json-encoded dictionary of values to be set as package attributes",
    )

    module.add_argument("module_file", metavar="<module file>")
    module.add_argument("command", metavar="<command>")
    module.add_argument(
        "--module-func", help="name of function to execute if command is execute"
    )
    module.add_argument("--args", help="arguments to pass to the function")
    module.add_argument(
        "--initial-state",
        help="A json-encoded dictionary of values to be set as package attributes",
    )

    pybuild.set_defaults(func=wrapper_pybuild)
    module.set_defaults(func=wrapper_module)

    args = parser.parse_args()

    if len(sys.argv) == 1:
        parser.print_help()

    env.ALLOW_LOAD_ERROR = False
    env.SANDBOX = True

    if args.debug:
        env.DEBUG = True

    args.verbose = None
    args.quiet = None
    init_logger(args)
    # Set logging level manually. Args doesn't contain any of --verbose or --quiet
    if args.verbosity is not None:
        logging.root.setLevel(args.verbosity)

    args.func(args)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Displays filesystem conflicts between mods
"""

from portmod.l10n import l10n
from portmod.portmod import file_conflicts
from portmod.vfs import get_vfs_dirs


def add_conflicts_parser(subparsers, parents):
    """
    Main executable for openmw-conflicts executable
    """
    conflicts_parser = subparsers.add_parser(
        "conflict-ui", help=l10n("conflict-ui-help"), parents=parents
    )

    def conflicts_main(args):
        mod_dirs = get_vfs_dirs()

        file_conflicts(mod_dirs, ["txt", "md"])

    conflicts_parser.set_defaults(func=conflicts_main)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
CLI to select various configuration options

Currently just profiles
"""

import os
from logging import info
from typing import List

from portmod.colour import bright, green, lblue
from portmod.config import get_config, set_config_value
from portmod.globals import env
from portmod.l10n import l10n
from portmod.modules import add_parsers
from portmod.news import add_news_parsers
from portmod.repo import RemoteRepo, has_repo
from portmod.repo.metadata import get_profiles
from portmod.repos import add_repo, get_repos, parse_remote_repos
from portmod.sync import sync

from .error import InputException


def display_list(items, notes, selected):
    padding = len(str(len(items)))
    for index, (item, note) in enumerate(zip(items, notes)):
        selected_str = ""
        if index in selected:
            selected_str = lblue("*")
        index_str = bright("[" + str(index) + "]")
        print(
            f'  {index_str} {" " * (padding - len(str(index))) + item} ({note})',
            selected_str,
        )


def add_profile_parsers(subparsers, parents):
    profile = subparsers.add_parser(
        "profile", help=l10n("profile-help"), parents=parents
    )
    profile_subparsers = profile.add_subparsers()
    profile_list = profile_subparsers.add_parser("list", help=l10n("profile-list-help"))
    profile_set = profile_subparsers.add_parser("set", help=l10n("profile-set-help"))
    profile_set.add_argument(
        "index", metavar=l10n("number-placeholder"), help=l10n("profile-number-help")
    )
    profile_show = profile_subparsers.add_parser("show", help=l10n("profile-show-help"))

    def get_profile():
        linkpath = os.path.join(env.prefix().PORTMOD_CONFIG_DIR, "profile")
        if os.path.exists(linkpath) and os.path.islink(linkpath):
            return os.readlink(linkpath).split("profiles")[-1].lstrip(os.path.sep)
        return None

    def list_func(args):
        profiles = get_profiles()
        print(bright(green(l10n("profile-available"))))
        display_list(
            [profile for _, profile, _ in profiles],
            [stability for _, _, stability in profiles],
            [
                index
                for index, (_, profile, _) in enumerate(profiles)
                if profile == get_profile()
            ],
        )

    def set_func(args):
        os.makedirs(env.prefix().PORTMOD_CONFIG_DIR, exist_ok=True)
        linkpath = os.path.join(env.prefix().PORTMOD_CONFIG_DIR, "profile")
        if os.path.exists(linkpath):
            os.unlink(linkpath)
        (path, _, _) = get_profiles()[int(args.index)]
        os.symlink(path, linkpath)

    def show_func(args):
        linkpath = os.path.join(env.prefix().PORTMOD_CONFIG_DIR, "profile")
        print(bright(green(l10n("profile-current-symlink", path=linkpath))))
        print(
            "  "
            + bright(os.readlink(linkpath).split("profiles")[-1].lstrip(os.path.sep))
        )

    def profile_help(args):
        profile.print_help()

    profile.set_defaults(func=profile_help)
    profile_list.set_defaults(func=list_func)
    profile_set.set_defaults(func=set_func)
    profile_show.set_defaults(func=show_func)


def add_repo_parser(subparsers, parents):
    repo = subparsers.add_parser("repo", help=l10n("repo-help"), parents=parents)
    repo_subparsers = repo.add_subparsers()
    repo_list = repo_subparsers.add_parser("list", help=l10n("repo-list-help"))
    repo_add = repo_subparsers.add_parser("add", help=l10n("repo-add-help"))
    repo_add.add_argument(
        "repo", metavar=l10n("repo-placeholder"), help=l10n("repo-identifier-help")
    )
    repo_remove = repo_subparsers.add_parser("remove", help=l10n("repo-remove-help"))
    repo_remove.add_argument(
        "repo", metavar=l10n("repo-placeholder"), help=l10n("repo-identifier-help")
    )

    def get_repos_list() -> List[RemoteRepo]:
        repos = []
        for repo in env.REPOS:
            conf = os.path.join(repo.location, "metadata", "repos.cfg")
            if os.path.exists(conf):
                repos.extend(parse_remote_repos(conf))
        return repos

    def list_func(args):
        repos = get_repos_list()
        print(bright(green(l10n("repos-available"))))
        display_list(
            [bright(repo.name) + ": " + repo.description for repo in repos],
            [repo.quality + " " + lblue(repo.sync_uri) for repo in repos],
            [index for index, repo in enumerate(repos) if has_repo(repo.name)],
        )

    def get_repo_name(value: str) -> str:
        repos = get_repos_list()
        if value.isdigit():
            repo_name = repos[int(value)].name
        else:
            if any(repo.name == value for repo in repos):
                repo_name = value
            else:
                raise InputException(l10n("repo-does-not-exist", name=value))

        return repo_name

    def add_func(args):
        enabled_repos = get_config()["REPOS"]
        repo_name = get_repo_name(args.repo)

        if repo_name and repo_name not in enabled_repos:
            enabled_repos.add(repo_name)
            info(l10n("repo-adding", name=repo_name, conf=env.prefix().PORTMOD_CONFIG))

        set_config_value("REPOS", " ".join(sorted(enabled_repos)))

        repo = next(repo for repo in get_repos_list() if repo.name == repo_name)
        result = add_repo(repo)
        if result:
            sync([result])
        env.REPOS = get_repos()

    def remove_func(args):
        enabled_repos = get_config()["REPOS"]
        repo_name = get_repo_name(args.repo)

        if repo_name and repo_name in enabled_repos:
            enabled_repos.remove(repo_name)
            info(
                l10n("repo-removing", name=repo_name, conf=env.prefix().PORTMOD_CONFIG)
            )

        set_config_value("REPOS", " ".join(sorted(enabled_repos)))

    def repo_help(args):
        repo.print_help()

    repo.set_defaults(func=repo_help)
    repo_list.set_defaults(func=list_func)
    repo_add.set_defaults(func=add_func)
    repo_remove.set_defaults(func=remove_func)


def add_select_parser(subparsers, parents):
    """
    Adds the select subparser to the given subparsers
    """
    parser = subparsers.add_parser("select", help=l10n("select-help"), parents=parents)
    _subparsers = parser.add_subparsers()
    add_profile_parsers(_subparsers, parents)
    add_news_parsers(_subparsers, parents)
    add_repo_parser(_subparsers, parents)
    add_parsers(_subparsers, parents)
    parser.set_defaults(func=lambda args: parser.print_help())



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
The module accessible within pybuilds

Note that this module should not be imported outside of pybuild files
"""

import portmod.globals  # noqa  # pylint: disable=unused-import
from portmod.atom import version_gt  # noqa  # pylint: disable=unused-import
from portmod.fs.util import patch_dir  # noqa  # pylint: disable=unused-import
from portmod.masters import get_masters  # noqa  # pylint: disable=unused-import
from portmod.parsers.usestr import use_reduce  # noqa  # pylint: disable=unused-import
from portmod.vfs import find_file, list_dir  # noqa  # pylint: disable=unused-import

from .pybuild import (  # noqa  # pylint: disable=unused-import
    File,
    InstallDir,
    Pybuild1,
    apply_patch,
)

DOWNLOAD_DIR = portmod.globals.env.DOWNLOAD_DIR

__all__ = [
    "version_gt",
    "patch_dir",
    "get_masters",
    "use_reduce",
    "find_file",
    "list_dir",
    "File",
    "InstallDir",
    "Pybuild1",
    "apply_patch",
    "DOWNLOAD_DIR",
]



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests the mod selection system
"""

import pytest

from portmod.atom import Atom
from portmod.loader import load_pkg

from .env import setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    yield setup_env("test")
    tear_down_env()


def test_moved():
    """Tests that moved atoms are loaded when the old name is used"""
    pkgs = load_pkg(Atom("test2/test-old"))
    assert any(pkg.CPN == "test/test-new" for pkg in pkgs)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Atom parsing tests
"""

import pytest

from portmod.atom import Atom, InvalidAtom


def test_invalid():
    with pytest.raises(InvalidAtom):
        Atom("foo/bar baz")


def test_valid():
    Atom("foo/bar-baz-1.0.20.3-r123::foo[-a,b?,!c?,c]")


def test_complex_version():
    Atom("foo/bar-1.0a_alpha")
    Atom("foo/bar-1.0a_alpha12")
    Atom("foo/bar-1.0a_pre1")
    Atom("foo/bar-1.0a_beta2")
    Atom("foo/bar-1.0a_rc1")
    Atom("foo/bar-1.0a_p1")


def test_nodots():
    assert Atom("foo/bar-1").PV == "1"



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

from typing import List, Union

from portmod.atom import Atom
from portmod.parsers.usestr import parse_usestr


def token_conflicts(token1: str, token2: str) -> bool:
    """
    Returns true if and only if two tokens, which use minus-format (e.g. -foo)
    to indicate a disabled token, conflict. E.g. foo and -foo
    """
    return (
        token1.lstrip("-") == token2.lstrip("-")
        and token1[0] == "-"
        and token2[0] != "-"
        or token1[0] != "-"
        and token2[0] == "-"
    )


def negate(token: str) -> str:
    """Returns the negation of the given token"""
    if token.startswith("-"):
        return token.lstrip("-")
    return "-" + token


def expand_use_conditionals(tokens: List[str]) -> List[Union[str, List]]:
    """Expands any conditional use dependencies in the token tree"""
    result = []
    for token in tokens:
        if isinstance(token, list):
            result.append(expand_use_conditionals(token))
        elif isinstance(token, Atom) and token.USE:
            for flag in token.USE:
                stripped = token.strip_use()
                sflag = flag.rstrip("?=").lstrip("!")
                if flag.endswith("?") and not flag.startswith("!"):
                    result += parse_usestr(
                        f"{sflag}? ( {stripped}[{sflag}] ) !{sflag}? ( {stripped} )",
                        Atom,
                    )
                elif flag.endswith("?") and flag.startswith("!"):
                    result += parse_usestr(
                        f"{sflag}? ( {stripped} ) !{sflag}? ( {stripped}[-{sflag}] )",
                        Atom,
                    )
                elif flag.endswith("=") and not flag.startswith("!"):
                    result += parse_usestr(
                        f"{sflag}? ( {stripped}[{sflag}] ) !{sflag}? ( {stripped}[-{sflag}] )",
                        Atom,
                    )
                elif flag.endswith("=") and flag.startswith("!"):
                    result += parse_usestr(
                        f"{sflag}? ( {stripped}[-{sflag}] ) !{sflag}? ( {stripped}[{sflag}] )",
                        Atom,
                    )
                else:
                    result.append(Atom(stripped + f"[{flag}]"))
        else:
            result.append(token)
    return result



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import os
import shutil
from logging import error, info, warning
from typing import Iterable

from packaging import version

from portmod.globals import env, get_version
from portmod.l10n import l10n
from portmod.news import update_news
from portmod.repo import Repo
from portmod.repos import get_local_repos


def sync(repos: Iterable[Repo]):
    # Slow imports
    import git

    for repo in repos:
        if repo.auto_sync and repo.sync_type == "git":

            if os.path.exists(repo.location):
                info(l10n("syncing-repo", repo=repo.name))
                gitrepo = git.Repo.init(repo.location)
                current = gitrepo.head.commit

                # Remote location has changed. Update gitrepo to match
                if gitrepo.remotes.origin.url != repo.sync_uri:
                    gitrepo.remotes.origin.set_url(repo.sync_uri)

                gitrepo.remotes.origin.pull(rebase=True)

                for diff in current.diff("HEAD"):
                    if diff.renamed_file:
                        print(
                            "{} {} -> {}".format(
                                diff.change_type, diff.a_path, diff.b_path
                            )
                        )
                    if diff.deleted_file:
                        print("{} {}".format(diff.change_type, diff.a_path))
                        if diff.a_path.endswith(".pybuild"):
                            # Remove from pybuild cache
                            parts = diff.a_path.split("/")
                            category = parts[0]
                            file = parts[-1].lstrip(".pybuild")
                            path = os.path.join(
                                env.PYBUILD_CACHE_DIR, repo.name, category, file
                            )
                            if os.path.exists(path):
                                os.remove(path)
                    else:
                        print("{} {}".format(diff.change_type, diff.b_path))

                tags = []
                for tag in gitrepo.tags:
                    # Valid tags must have the tag commit be the merge base
                    # A merge base further back indicates a branch point
                    if tag.name.startswith("portmod_v"):
                        base = gitrepo.merge_base(gitrepo.head.commit, tag.commit)
                        if base and base[0] == tag.commit:
                            tags.append(tag)

                this_version = version.parse(get_version())
                newest = max(
                    [version.parse(tag.name.lstrip("portmod_v")) for tag in tags]
                    + [this_version]
                )
                if newest != this_version and not env.TESTING:
                    warning(l10n("update-message"))
                    warning(l10n("current-version", version=this_version))
                    warning(l10n("new-version", version=newest))
                info(l10n("done-syncing-repo", repo=repo.name))
            else:
                git.Repo.clone_from(repo.sync_uri, repo.location)
                info(l10n("initialized-repository", repo=repo.name))
        elif repo.auto_sync:
            error(
                l10n(
                    "invalid-sync-type",
                    type=repo.sync_type,
                    repo=repo.name,
                    supported="git",
                )
            )

    if os.path.exists(env.PYBUILD_CACHE_DIR):
        local = get_local_repos()
        for repo_name in os.listdir(env.PYBUILD_CACHE_DIR):
            path = os.path.join(env.PYBUILD_CACHE_DIR, repo_name)
            if repo_name not in local:
                print(l10n("cache-cleanup", repo=repo_name))
                shutil.rmtree(path)

    update_news()



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import configparser
import getpass
import os
import re
from functools import lru_cache
from logging import info, warning
from stat import S_IWRITE
from typing import Mapping

from .globals import env
from .l10n import l10n
from .repo.metadata import get_archs
from .repos import get_local_repos


class InvalidPrefix(RuntimeError):
    def __init__(self, prefix):
        self.prefix = prefix

    def __str__(self):
        return l10n("invalid-prefix", prefix=self.prefix)


class PrefixExistsError(RuntimeError):
    def __init__(self, prefix):
        self.prefix = prefix

    def __str__(self):
        return l10n("prefix-exists", prefix=self.prefix)


@lru_cache()
def get_prefixes() -> Mapping[str, str]:
    """Returns a mapping of prefixes to their architectures"""
    prefixes = {}
    if os.path.exists(env.PREFIX_FILE):
        with open(env.PREFIX_FILE, "r") as file:
            for line in file.readlines():
                line = re.sub("#.*", "", line)
                if line:
                    prefix, arch = line.split()
                    prefixes[prefix] = arch

    return prefixes


def add_prefix(prefix: str, arch: str):
    """Adds a new prefix"""

    invalid_prefixes = {"news", "sync", "mirror"}

    if prefix in invalid_prefixes:
        raise InvalidPrefix(prefix)

    if prefix in get_prefixes():
        raise PrefixExistsError(prefix)

    arch_options = set()
    for repo in get_local_repos().values():
        arch_options |= get_archs(repo.location)

    if arch not in arch_options:
        warning(l10n("unknown-arch", arch=arch))

    # Add new prefix to the prefix file
    if os.path.exists(env.PREFIX_FILE):
        stat = os.stat(env.PREFIX_FILE)
        os.chmod(env.PREFIX_FILE, stat.st_mode | S_IWRITE, follow_symlinks=True)

    os.makedirs(os.path.dirname(env.PREFIX_FILE), exist_ok=True)

    with open(env.PREFIX_FILE, "a") as file:
        print(prefix, arch, file=file)

    stat = os.stat(env.PREFIX_FILE)
    os.chmod(env.PREFIX_FILE, stat.st_mode - S_IWRITE, follow_symlinks=True)

    get_prefixes.cache_clear()

    env.set_prefix(prefix)

    # Ensure that INSTALLED_DB exists
    if not os.path.exists(env.prefix().INSTALLED_DB):
        import git

        # Initialize as git repo
        os.makedirs(env.prefix().INSTALLED_DB)
        gitrepo = git.Repo.init(env.prefix().INSTALLED_DB)
        # This repository is for local purposes only.
        # We don't want to worry about prompts for the user's gpg key password
        localconfig = gitrepo.config_writer()
        localconfig.set_value("commit", "gpgsign", False)
        USER = getpass.getuser()

        try:
            # May throw TypeError if GitPython<3.0.5
            globalconfig = git.config.GitConfigParser()
            globalconfig.get_value("user", "name")
            globalconfig.get_value("user", "email")
        except (TypeError, configparser.NoOptionError, configparser.NoSectionError):
            # Set the user name and email if they aren't in a global config
            localconfig.set_value("user", "name", f"{USER}")
            localconfig.set_value("user", "email", f"{USER}@example.com")

        localconfig.release()

    get_prefixes.cache_clear()
    info(l10n("initialized-prefix", prefix=prefix))



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Module for directly loading pybuild files.
These functions should not be called directly.
See portmod.loader for functions to load pybuilds safely using a sandbox.
"""

import ast
import glob
import importlib
import importlib.util
import os
import sys
from functools import lru_cache
from logging import warning
from types import SimpleNamespace
from typing import Any, Callable, Dict, Generator, Optional, cast

from RestrictedPython import (
    RestrictingNodeTransformer,
    compile_restricted_exec,
    limited_builtins,
    safe_globals,
)
from RestrictedPython.Eval import default_guarded_getitem, default_guarded_getiter
from RestrictedPython.Guards import (
    guarded_iter_unpack_sequence,
    guarded_unpack_sequence,
    safer_getattr,
)

from portmod.atom import Atom, FQAtom, QualifiedAtom, VAtom, atom_sat
from portmod.functools import install_cache
from portmod.globals import env
from portmod.l10n import l10n
from portmod.lock import vdb_lock
from portmod.parsers.manifest import Manifest
from portmod.pybuild import FullInstalledPybuild, FullPybuild
from portmod.repo.metadata import get_categories
from portmod.util import get_max_version

from . import get_repo, get_repo_name
from .metadata import get_masters
from .updates import get_moved

WHITELISTED_IMPORTS = {
    "filecmp",
    "os",
    "sys",
    "shutil",
    "os.path",
    "chardet",
    "pybuild",
    "pybuild.info",
    "pybuild.winreg",
    "re",
    "csv",
    "json",
    "typing",
    "fnmatch",
    "collections",
}


@lru_cache()
def _import_common(
    name: str,
    installed: bool,
    load_function: Callable,
    repo: Optional[str] = None,
) -> SimpleNamespace:
    """
    args:
        name: The import name as an absolute import path
        installed: Whether or not the package calling this is an installed package
        load_function: The function taking a file path and a keyword argument installed,
            indicating the installed status of the file to be loaded, to be used to load
            the common module

    returns:
        The Module as a SimpleNamespace
    """
    if len(name.split(".")) > 2:
        raise Exception(f"Invalid package {name}")
    _, module_name = name.split(".")
    base_atom = Atom(f"common/{module_name}")
    if installed:
        path = find_installed_path(base_atom)

    if installed and path:
        return SimpleNamespace(**load_function(path, installed=True, repo=None))
    else:
        versions = {}
        for file in _iterate_pybuilds(base_atom, repo_name=repo):
            atom, _ = os.path.splitext(file)
            versions[VAtom("common/" + os.path.basename(atom)).PVR] = file

        max_version = get_max_version(versions.keys())
        if not max_version:
            raise Exception(f"Could not find package {name}")
        return SimpleNamespace(
            **load_function(versions[max_version], installed=False, repo=repo)
        )


def _import(installed: bool, repo: Optional[str] = None):
    def import_fn(name, globs, loc, fromlist, level):
        if name.startswith("common."):
            return _import_common(name, installed, __load_module, repo)
        if name not in WHITELISTED_IMPORTS:
            raise RuntimeError("Importing from {name} is not allowed")
        return __import__(name, globs, loc, fromlist, level)

    return import_fn


# Default implementation to handle invalid pybuilds
class Package:
    def __init__(self):
        raise Exception("Package is not defined")


def default_write_guard(ob):
    """Noop write guard"""
    return ob


def safer_hasattr(obj, name):
    """
    Version of hasattr implemented using safet_getattr

    This doesn't really provide any extra security, but does mean that
    str.format, and attributes beginning with underscores, which are
    blocked by safer_getattr, return False rather than True
    """
    try:
        safer_getattr(obj, name)
    except (NotImplementedError, AttributeError):
        return False
    return True


def default_apply(func, *args, **kwargs):
    return func(*args, **kwargs)


SAFE_GLOBALS: Dict[str, Any] = safe_globals
SANDBOX_GLOBALS: Dict[str, Any] = safe_globals.copy()
SANDBOX_GLOBALS["__builtins__"] = safe_globals["__builtins__"].copy()
SANDBOX_GLOBALS.update({"Package": Package})


class PrintWrapper:
    def __init__(self, _getattr_=None):
        self.txt = []
        self._getattr_ = _getattr_

    def write(self, text):
        self.txt.append(text)

    def __call__(self):
        return "".join(self.txt)

    def _call_print(self, *objects, **kwargs):
        if kwargs.get("file", None) is None:
            kwargs["file"] = sys.stdout
        else:
            self._getattr_(kwargs["file"], "write")
        print(*objects, **kwargs)


SANDBOX_GLOBALS["__builtins__"].update(
    {
        "FileNotFoundError": FileNotFoundError,
        "__metaclass__": type,
        "_apply_": default_apply,
        "_getattr_": safer_getattr,
        "_getitem_": default_guarded_getitem,
        "_getiter_": default_guarded_getiter,
        "_iter_unpack_sequence_": guarded_iter_unpack_sequence,
        "_print_": PrintWrapper,
        "_unpack_sequence_": guarded_unpack_sequence,
        "_write_": default_write_guard,
        "all": all,
        "any": any,
        "dict": dict,
        "enumerate": enumerate,
        "filter": filter,
        "frozenset": frozenset,
        "getattr": safer_getattr,
        "hasattr": safer_hasattr,
        "iter": iter,
        "map": map,
        "max": max,
        "min": min,
        "next": next,
        "open": open,
        "reversed": reversed,
        "set": set,
        "sorted": sorted,
        "sum": sum,
        "super": super,
    }
)
SANDBOX_GLOBALS["__builtins__"].update(limited_builtins)


class Policy(RestrictingNodeTransformer):
    def visit_JoinedStr(self, node):
        return self.node_contents_visit(node)

    def visit_FormattedValue(self, node):
        return self.node_contents_visit(node)

    def visit_AnnAssign(self, node):
        return self.node_contents_visit(node)

    def visit_AugAssign(self, node):
        return self.node_contents_visit(node)

    def visit_ImportFrom(self, node):
        if (
            node.module in WHITELISTED_IMPORTS
            or node.module.startswith("pybuild")
            and node.level == 0
        ):
            try:
                module = importlib.import_module(node.module)
                for name in node.names:
                    if isinstance(module.__dict__.get(name.name), type(importlib)):
                        self.error(
                            node, "Importing modules from other modules is forbidden"
                        )
            except ModuleNotFoundError:
                pass

        return RestrictingNodeTransformer.visit_ImportFrom(self, node)

    def visit_FunctionDef(self, node):
        node = RestrictingNodeTransformer.visit_FunctionDef(self, node)
        if node.name == "__init__":
            newnode = ast.parse("super().__init__()").body[0]
            newnode.lineno = node.lineno
            newnode.col_offset = node.col_offset
            node.body.insert(0, newnode)
        return node


def restricted_load(code, filepath: str, _globals: Dict[str, Any]):
    if sys.platform == "win32":
        code = code.replace("\\", "\\\\")
    byte_code, errors, warnings, names = compile_restricted_exec(
        code, filename=filepath, policy=Policy
    )
    if errors:
        raise SyntaxError(errors)
    seen: Dict[str, str] = {}
    for message in [seen.setdefault(x, x) for x in warnings if x not in seen]:
        if not message.endswith("Prints, but never reads 'printed' variable."):
            warning(f"In file {filepath}: {message}")
    exec(byte_code, _globals, _globals)


@install_cache
def get_atom_from_path(path: str) -> FQAtom:
    repopath, filename = os.path.split(os.path.abspath(os.path.normpath(path)))
    atom, _ = os.path.splitext(filename)
    repopath, _ = os.path.split(repopath)
    repopath, C = os.path.split(repopath)
    try:
        repo_name = get_repo_name(repopath)
    except FileNotFoundError as e:
        path = os.path.join(os.path.dirname(path), "REPO")
        if os.path.exists(path):
            with open(path, "r") as file:
                repo_name = file.read().strip() + "::installed"
        else:
            raise e
    return FQAtom(f"{C}/{atom}::{repo_name}")


def __load_module(
    path: str, *, installed=False, repo: Optional[str] = None
) -> Dict[str, Any]:
    filename, _ = os.path.splitext(os.path.basename(path))

    with open(path, "r", encoding="utf-8") as file:
        code = file.read()
        tmp_globals = SANDBOX_GLOBALS.copy()
        tmp_globals["__builtins__"]["__import__"] = _import(
            installed=installed, repo=repo
        )
        tmp_globals["__name__"] = filename
        restricted_load(code, path, tmp_globals)
        for _global in tmp_globals.values():
            # Note: we can't use hasattr here, as we want to set the attribute for this classmethod
            if isinstance(_global, type) and "__file__" not in _global.__dict__:
                try:
                    setattr(_global, "__file__", path)
                except:  # noqa
                    ...

    return tmp_globals


def __load_file(path: str, *, installed=False) -> FullPybuild:
    """
    Loads a pybuild file

    :param path: Path of the pybuild file
    """
    repo = get_repo_name(path) if not installed else None
    module = __load_module(path, installed=installed, repo=repo)
    module["Package"].__pybuild__ = path
    mod = module["Package"]()
    mod.FILE = os.path.abspath(path)
    mod.INSTALLED = False

    if not installed:
        # determine common dependencies
        def find_common_imports(file: str):
            depends = []
            with open(file, "r", encoding="utf-8") as fp:
                tree = ast.parse(fp.read())

            def find_imports(tree: ast.AST):
                if isinstance(tree, ast.Module):
                    for statement in tree.body:
                        if isinstance(statement, ast.Import):
                            for alias in statement.names:
                                if alias.name.startswith("common."):
                                    depends.append(alias.name.replace(".", "/"))
                        elif isinstance(statement, ast.ImportFrom):
                            if statement.module and statement.module.startswith(
                                "common."
                            ):
                                depends.append(statement.module.replace(".", "/"))

            # Globals
            find_imports(tree)
            # TODO: Inline imports in functions?
            return depends

        mod.RDEPEND = " ".join([mod.RDEPEND] + find_common_imports(mod.FILE))
    return cast(FullPybuild, mod)


@vdb_lock()
def __load_installed(file: str) -> FullInstalledPybuild:
    """
    Loads an installed pybuild

    :param file: Path of the pybuild file
    """
    mod = cast(FullInstalledPybuild, __load_file(file, installed=True))
    mod.INSTALLED = True
    parent = os.path.dirname(file)

    def read_file(name: str) -> Optional[str]:
        if os.path.exists(os.path.join(parent, name)):
            with open(os.path.join(parent, name), "r") as repo_file:
                return repo_file.read().strip()
        return None

    repo = read_file("REPO")
    if not repo:
        raise Exception(
            f"Internal Error: Installed package in file {file}"
            "has no repository identifier"
        )
    mod.REPO = repo
    mod.INSTALLED_USE = set((read_file("USE") or "").split())
    mod.RDEPEND = read_file("RDEPEND") or mod.RDEPEND
    mod.INSTALLED_REBUILD_FILES = None
    path = os.path.join(parent, "REBUILD_FILES")
    if os.path.exists(path):
        try:
            mod.INSTALLED_REBUILD_FILES = Manifest(path)
        except ValueError:
            warning(f"Failed to load {path}")
            pass
    return mod


def _iterate_installed() -> Generator[str, None, None]:
    repo = env.prefix().INSTALLED_DB

    for path in glob.glob(os.path.join(repo, "*", "*", "*.pybuild")):
        yield path


def find_installed_path(atom: Atom) -> Optional[str]:
    repo_path = env.prefix().INSTALLED_DB

    # Handle renames:
    # Check all possible renames for this atom (all repos), and they are only valid if
    # the package comes from the repo matching the rename entry
    for repo in env.REPOS:
        moved = get_moved(repo)
        if atom.CPN in moved:
            moved_atom = QualifiedAtom(moved[atom.CPN])
            path = os.path.join(repo_path, moved_atom.C, moved_atom.PN)
            if os.path.exists(path):
                with open(os.path.join(path, "REPO"), "r") as file:
                    if repo.name == file.read().strip():
                        atom = moved_atom
                        break

    if atom.C:
        path = os.path.join(repo_path, atom.C, atom.PN)
        if os.path.exists(path):
            results = glob.glob(os.path.join(path, "*.pybuild"))
            assert len(results) == 1
            return results[0]
    else:
        for dirname in glob.glob(os.path.join(repo_path, "*")):
            path = os.path.join(repo_path, dirname, atom.PN)
            if os.path.exists(path):
                results = glob.glob(os.path.join(path, "*.pybuild"))
                assert len(results) == 1
                return results[0]
    return None


def _iterate_pybuilds(
    atom: Optional[Atom] = None, repo_name: Optional[str] = None
) -> Generator[str, None, None]:
    path = None
    repos = []
    if env.PREFIX_NAME:
        repos = env.prefix().REPOS
    else:
        repos = env.REPOS

    if repo_name is not None:
        repo = get_repo(repo_name)
        repos = [repo]
        for master in get_masters(repo.location):
            yield from _iterate_pybuilds(atom, master.name)

    def valid_atom(desired_atom: Atom, other: Atom):
        if isinstance(desired_atom, FQAtom):
            return desired_atom == other
        else:
            return atom_sat(other, desired_atom)

    def try_move_atom(moved: Dict[str, str], atom: QualifiedAtom, repo):
        true_atom = atom
        if atom.CPN in moved:
            true_atom = QualifiedAtom(moved[atom.CPN])
            if atom.PVR:
                true_atom = QualifiedAtom(f"{true_atom}-{atom.PVR}")
        return true_atom

    for repo in repos:
        if not os.path.exists(repo.location):
            warning(
                l10n(
                    "repo-does-not-exist-warning",
                    name=repo.name,
                    path=repo.location,
                    command="omwmerge --sync",
                )
            )

        if atom:
            moved = get_moved(repo)
            if atom.C:
                true_atom = try_move_atom(moved, cast(QualifiedAtom, atom), repo)
                path = os.path.join(repo.location, true_atom.C, true_atom.PN)
                if path is not None and os.path.exists(path):
                    for file in glob.glob(os.path.join(path, "*.pybuild")):
                        if valid_atom(true_atom, get_atom_from_path(file)):
                            yield file
            else:
                for category in get_categories(repo.location):
                    true_atom = try_move_atom(
                        moved, QualifiedAtom(category + "/" + atom), repo
                    )
                    path = os.path.join(repo.location, category, true_atom.PN)

                    if path is not None and os.path.exists(path):
                        for file in glob.glob(os.path.join(path, "*.pybuild")):
                            if valid_atom(true_atom, get_atom_from_path(file)):
                                yield file
        else:
            for file in glob.glob(os.path.join(repo.location, "*", "*", "*.pybuild")):
                yield file


def pkg_exists(atom: Atom, *, repo_name: Optional[str] = None) -> bool:
    """Returns true if a package with the given atom can be found in the given repository"""
    return next(_iterate_pybuilds(atom, repo_name), None) is not None



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import shutil
from collections.abc import Mapping

import pytest

from portmod.repo.loader import __load_file

from .env import setup_env, tear_down_env
from .test_loader import TMP_REPO, create_pybuild


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    yield dictionary
    tear_down_env()
    shutil.rmtree(TMP_REPO)


def test_import_side_effects():
    """
    Tests for side effects when importing whitelisted python modules

    This specific instance is obscure, and it's not clear why it occurs, however
    avoiding using our custom import code for whitelisted modules avoids this causing
    an exception. See #135.
    """
    file = """
from pybuild import Pybuild1
import typing

class Package(Pybuild1):
    NAME="Test"
    DESC="Test"
    LICENSE="GPL-3"
"""

    __load_file(create_pybuild(file))

    isinstance([], Mapping)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Tests the setup systems
"""

import sys
from contextlib import redirect_stdout
from io import StringIO

import pytest

from portmod._cli.main import main
from portmod.globals import env
from portmod.prefix import get_prefixes
from portmod.repos import get_local_repos

from .env import setup_env, tear_down_env


@pytest.fixture(autouse=True)
def setup():
    """
    Sets up test repo for querying
    """
    yield setup_env("test")
    tear_down_env()


def test_init_prefix():
    """Tests prefix creation"""
    sys.argv = ["portmod", "init", "test2", "test"]
    main()

    assert get_prefixes()["test2"] == "test"


def test_add_repo():
    """Tests adding repositories automatically"""
    stringio = StringIO()
    with redirect_stdout(stringio):
        sys.argv = ["portmod", "test", "select", "repo", "list"]
        main()
    assert "blank" in stringio.getvalue()

    sys.argv = ["portmod", "test", "select", "repo", "add", "blank"]
    main()

    assert any(repo.name == "blank" for repo in env.REPOS)
    assert "blank" in get_local_repos()



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Helper functions for interacting with package licenses
"""

from functools import lru_cache
from typing import List, Union

from portmod.config import get_config
from portmod.config.use import get_use
from portmod.parsers.usestr import use_reduce
from portmod.pybuild import Pybuild
from portmod.repo import get_repo, get_repo_root
from portmod.repo.metadata import get_license_groups


def is_license_accepted(mod: Pybuild) -> bool:
    """
    Returns true if the mod's license(s) are accepted by the user's configuration

    For a license to be accepted, it must be both listed, either explicitly,
    part of a group, or with the * wildcard  and it must not be blacklisted
    by a license or license group prefixed by a '-'
    """
    if mod.INSTALLED:
        license_groups = get_license_groups(get_repo(mod.REPO).location)
    else:
        license_groups = get_license_groups(get_repo_root(mod.FILE))

    ACCEPT_LICENSE = get_config()["ACCEPT_LICENSE"]

    def accepted(group: Union[str, List]) -> bool:
        if not group:
            return True

        if isinstance(group, str):
            allowed = False
            # Check if license is allowed by anything in ACCEPT_LICENSE
            for license in ACCEPT_LICENSE:
                if license.startswith("-") and (
                    license == group
                    or (license[1] == "@" and group in license_groups[license[2:]])
                ):
                    # not allowed if matched by this
                    return False
                if license == "*":
                    allowed = True
                if license.startswith("@") and group in license_groups[license[1:]]:
                    allowed = True
            return allowed
        if group[0] == "||":
            return any(accepted(license) for license in group)

        return all(accepted(license) for license in group)

    enabled, disabled = get_use(mod)
    return accepted(use_reduce(mod.LICENSE, enabled, disabled, opconvert=True))

    # TODO: implement package-specific license acceptance via package.license config file


@lru_cache()
def has_eula(package: Pybuild) -> bool:
    groups = get_license_groups(get_repo_root(package.FILE))
    # FIXME: This should be reworked.
    # For one thing, this doesn't currently handle || operators
    return any(
        license_name in groups.get("EULA", set())
        for license_name in use_reduce(
            package.LICENSE, get_use(package)[0], get_use(package)[1], flat=True
        )
    )



#!/usr/bin/env python

# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3


import os

from setuptools import find_packages, setup
from setuptools_rust import Binding, RustExtension

with open(os.path.join(os.path.dirname(__file__), "README.md"), "r") as file:
    long_description = file.read()


setup(
    name="portmod",
    author="Portmod Authors",
    author_email="incoming+portmod-portmod-9660349-issue-@incoming.gitlab.com",
    description="A CLI package manager for mods",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GPLv3",
    url="https://gitlab.com/portmod/portmod",
    download_url="https://gitlab.com/portmod/portmod/-/releases",
    packages=find_packages(exclude=["*.test", "*.test.*", "test.*", "test"]),
    rust_extensions=[
        RustExtension("portmod.portmod", binding=Binding.PyO3, strip=True)
    ],
    zip_safe=False,
    entry_points=(
        {
            "console_scripts": [
                "inquisitor = portmod._cli.inquisitor:main",
                "portmod= portmod._cli.main:main",
            ]
        }
    ),
    python_requires=">=3.6",
    install_requires=[
        "patool",
        "colorama",
        "appdirs",
        "GitPython",
        "progressbar2",
        'pywin32; platform_system == "Windows"',
        "RestrictedPython>=4.0",
        "redbaron",
        'python-sat; platform_system != "Windows"',
        'python-sat>=0.1.5.dev12; platform_system == "Windows"',
        "requests",
        "chardet",
        'importlib_metadata; python_version < "3.8"',
        "packaging",
        "fasteners>=0.16",
    ],
    setup_requires=["setuptools_scm", 'wheel; platform_system == "Windows"'],
    use_scm_version={"write_to": "portmod/_version.py"},
    extras_require={
        "dev": ["black", "flake8", "pylint", "isort", "mypy"],
        "test": ["pytest", "pytest-cov", "setuptools_scm"],
        "benchmark": ["pytest-benchmark"],
        "bash": ["argcomplete"],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Games/Entertainment",
        "Topic :: System :: Software Distribution",
    ],
)



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import io
import logging
import os
import shutil
from logging import error, info, warning
from typing import Iterable, Optional

from .atom import Atom, QualifiedAtom, atom_sat
from .colour import bright, green, lblue, lgreen, red, yellow
from .config import get_config
from .config.sets import BUILTIN_SETS, add_set, get_set, remove_set
from .config.use import add_use
from .deps import resolve
from .download import fetchable, find_download, is_downloaded
from .globals import env
from .l10n import l10n
from .loader import (
    AmbiguousAtom,
    _sandbox_execute_pybuild,
    load_all_installed,
    load_installed_pkg,
    load_pkg,
)
from .modules import clear_module_updates, require_module_updates, update_modules
from .package import install_pkg, remove_pkg
from .perms import Permissions
from .prompt import prompt_bool
from .rebuild import get_rebuild_manifest
from .repo.keywords import add_keyword
from .repo.profiles import get_system
from .transactions import (
    Delete,
    Transactions,
    UseDep,
    generate_transactions,
    print_transactions,
    sort_transactions,
)
from .tsort import CycleException
from .util import KeywordDep, LicenseDep, select_package
from .vfs import clear_vfs_sort, require_vfs_sort, sort_vfs, vfs_needs_sorting


def global_updates():
    """Performs updates to global configuration"""
    # Update module set.
    for atom in get_set("modules", parent_dir=env.prefix().PORTMOD_LOCAL_DIR):
        if not load_installed_pkg(Atom(atom)):
            # Note: safe because this doesn't modify the set returned by get_set
            remove_set("modules", atom, parent_dir=env.prefix().PORTMOD_LOCAL_DIR)

    for pkg in load_all_installed():
        if "module" in pkg.PROPERTIES:
            add_set("modules", pkg.CPN, parent_dir=env.prefix().PORTMOD_LOCAL_DIR)

    # Fix vfs ordering and update modules

    try:
        sort_vfs()
        clear_vfs_sort()
    except CycleException as e:
        error(f"{e}")

    update_modules()
    clear_module_updates()


def deselect(pkgs: Iterable[str], *, no_confirm: bool = False):
    all_to_remove = []

    for name in pkgs:
        atom = Atom(name)
        to_remove = []
        for mod in get_set("selected-packages"):
            if atom_sat(mod, atom):
                to_remove.append(mod)

        if len(to_remove) == 1:
            info(">>> " + l10n("remove-from-world", atom=green(to_remove[0])))
            all_to_remove.append(to_remove[0])
        elif len(to_remove) > 1:
            raise AmbiguousAtom(atom, to_remove)

    if not all_to_remove:
        print(">>> " + l10n("no-matching-world-atom"))
        return

    if no_confirm or prompt_bool(bright(l10n("remove-from-world-qn"))):
        for mod in all_to_remove:
            remove_set("selected-packages", mod)


def configure(
    atoms: Iterable[str],
    *,
    delete: bool = False,
    depclean: bool = False,
    auto_depclean: bool = False,
    no_confirm: bool = False,
    oneshot: bool = False,
    verbose: bool = False,
    update: bool = False,
    newuse: bool = False,
    noreplace: bool = False,
    nodeps: bool = False,
    deselect: Optional[bool] = None,
    select: Optional[bool] = None,
    deep: bool = False,
    emptytree: bool = False,
):
    # Slow import
    import git

    # Ensure that we always get the config before performing operations on packages
    # This way the config settings will be available as environment variables.
    get_config()

    targetlist = list(atoms)
    for modstr in targetlist:
        if modstr.startswith("@"):
            # Atom is actually a set. Load set instead
            targetlist.extend(get_set(modstr.replace("@", "")))
            continue

    to_remove = set()
    if delete or depclean:
        for modstr in targetlist:
            if modstr.startswith("@"):
                continue

            skip = False
            atom = Atom(modstr)
            for system_atom in get_system():
                if atom_sat(system_atom, atom):
                    warning(l10n("skipping-system-package", atom=system_atom))
                    skip = True
                    break

            if not skip:
                to_remove.add(atom)

    atomlist = [
        Atom(modstr)
        for modstr in targetlist
        if modstr not in to_remove and not modstr.startswith("@")
    ]

    selected = set(Atom(atom) for atom in atoms if not atom.startswith("@"))
    selected_sets = set(atom[1:] for atom in atoms if atom.startswith("@"))

    explicit = set(selected)
    for selected_set in selected_sets:
        explicit |= get_set(selected_set)

    if delete:
        # Do nothing. We don't care about deps
        transactions = Transactions()
        for atom in to_remove:
            ipkg = load_installed_pkg(atom)
            if not ipkg:
                raise Exception(l10n("not-installed", atom=atom))
            transactions.append(Delete(ipkg))
    elif nodeps:
        fqatoms = []
        for atom in atomlist:
            pkg, _ = select_package(load_pkg(atom))
            fqatoms.append(pkg.ATOM)

        selected_cpn = set()
        for atom in selected:
            pkg, _ = select_package(load_pkg(atom))
            selected_cpn.add(QualifiedAtom(pkg.CPN))

        transactions = generate_transactions(
            fqatoms,
            [],
            set() if oneshot or depclean else selected_cpn,
            [],
            noreplace=noreplace,
            emptytree=emptytree,
        )
    else:
        transactions = resolve(
            atomlist,
            to_remove,
            set() if noreplace else explicit,
            set() if oneshot or depclean else selected,
            set() if oneshot or depclean else selected_sets,
            deep=deep
            or (depclean and not to_remove),  # No argument depclean implies deep
            update=update,
            newuse=newuse,
            noreplace=noreplace or depclean or update or newuse,
            depclean=auto_depclean or depclean,
            emptytree=emptytree,
        )

    transactions = sort_transactions(transactions)

    # Inform user of changes
    if transactions.pkgs:
        # Don't print transaction list when in quiet mode and no-confirm is passed
        if not no_confirm or logging.root.level < logging.WARN:
            if delete or depclean:
                print(l10n("to-remove"))
            else:
                print(l10n("to-install"))
            print_transactions(transactions, verbose=verbose)
            print()
    elif vfs_needs_sorting() and not transactions.pkgs:
        global_updates()
        info(l10n("nothing-else-to-do"))
        return
    elif not transactions.pkgs:
        info(l10n("nothing-to-do"))
        return

    if transactions.config:
        keyword_changes = list(
            filter(lambda x: isinstance(x, KeywordDep), transactions.config)
        )
        license_changes = list(
            filter(lambda x: isinstance(x, LicenseDep), transactions.config)
        )
        use_changes = list(filter(lambda x: isinstance(x, UseDep), transactions.config))
        if keyword_changes:
            print(l10n("necessary-keyword-changes"))
            for keyword in keyword_changes:
                if keyword.keyword.startswith("*"):
                    c = red
                else:
                    c = yellow
                print("    {} {}".format(green(keyword.atom), c(keyword.keyword)))

            if no_confirm or prompt_bool(l10n("apply-changes?")):
                for keyword in keyword_changes:
                    add_keyword(keyword.atom, keyword.keyword)
            else:
                return

        if license_changes:
            # TODO: For EULA licenses, display the license and prompt the user to accept
            print(l10n("necessary-license-changes"))
            for license in license_changes:
                print("    {} {}".format(green(license.atom), license.license))
            return

        if use_changes:
            print(l10n("necessary-flag-changes"))
            for use in use_changes:
                if use.flag.startswith("-") and use.oldvalue == use.flag.lstrip("-"):
                    print(
                        "    {} {} # {}".format(
                            lblue(use.atom), red(use.flag), l10n("enabled-comment")
                        )
                    )
                elif not use.flag.startswith("-") and use.oldvalue == "-" + use.flag:
                    print(
                        "    {} {} # {}".format(
                            green(use.atom), red(use.flag), l10n("disabled-comment")
                        )
                    )
                else:
                    print("    {} {}".format(green(use.atom), red(use.flag)))
            if no_confirm or prompt_bool(l10n("apply-changes-qn")):
                for use in use_changes:
                    add_use(use.flag.lstrip("-"), use.atom, use.flag.startswith("-"))
            else:
                return

    def print_restricted_fetch(transactions: Transactions):
        # Check for restricted fetch packages and print their nofetch notices
        for trans in transactions.pkgs:
            if not isinstance(trans, Delete):
                can_fetch = fetchable(trans.pkg)
                to_fetch = [
                    source
                    for source in trans.pkg.get_default_sources()
                    if find_download(source) is None
                ]
                if set(to_fetch) - set(can_fetch) and not is_downloaded(trans.pkg):
                    print(green(l10n("fetch-instructions", atom=trans.pkg.ATOM)))
                    state = {
                        "UNFETCHED": to_fetch,
                        "A": trans.pkg.get_default_sources(),
                        "USE": trans.pkg.get_use()[0],
                    }
                    _sandbox_execute_pybuild(
                        trans.pkg.FILE,
                        "nofetch",
                        Permissions(),
                        init=state,
                    )
                    print()

    print_restricted_fetch(transactions)

    tmp_dir = env.TMP_DIR
    # If TMP_DIR doesn't exist, either use the parent, or if that doesn't exist,
    # just create it
    if not os.path.exists(env.TMP_DIR):
        if os.path.exists(os.path.dirname(env.TMP_DIR)):
            tmp_dir = os.path.dirname(env.TMP_DIR)
        else:
            os.makedirs(tmp_dir, exist_ok=True)
    tmp_space = shutil.disk_usage(tmp_dir).free

    for trans in transactions.pkgs:
        if not isinstance(trans, Delete) and "pkg_pretend" in trans.pkg.FUNCTIONS:
            # TODO: There are various variables that should be set on mod during pkg_pretend
            _sandbox_execute_pybuild(
                trans.pkg.FILE,
                "pretend",
                Permissions(),
                save_state=True,
            )

            for source in trans.pkg.get_default_sources():
                if source.size > 4 * tmp_space:
                    warning(
                        l10n(
                            "tmp-space-too-small",
                            dir=env.TMP_DIR,
                            free=tmp_space / 1024 / 1024,
                            size=source.size * 4 / 1024 / 1024,
                        )
                    )

    if not (no_confirm or prompt_bool(l10n("continue-qn"))):
        return

    err = None
    merged = Transactions()
    # Install (or remove) packages in order
    for trans in transactions.pkgs:
        if isinstance(trans, Delete):
            remove_pkg(trans.pkg)
            if deselect is None or deselect:
                if trans.pkg.CPN in get_set("selected-packages"):
                    info(">>> " + l10n("remove-from-world", atom=green(trans.pkg.CPN)))
                    remove_set("selected-packages", trans.pkg.CPN)
            merged.pkgs.append(trans)
        elif install_pkg(trans.pkg):
            if trans.pkg in transactions.new_selected and not oneshot:
                if trans.pkg.CPN not in get_set("selected-packages"):
                    info(">>> " + l10n("add-to-world", atom=green(trans.pkg.CPN)))
                    add_set("selected-packages", trans.pkg.CPN)
            merged.pkgs.append(trans)
        else:
            # Unable to install mod. Aborting installing remaining packages
            err = trans.pkg.ATOM
            break

        require_vfs_sort()
        require_module_updates()
        sort_vfs()
        clear_vfs_sort()

    for trans in merged.pkgs:
        pkg = trans.pkg
        warning_path = os.path.join(env.WARNINGS_DIR, pkg.ATOM.CPF)
        info_path = os.path.join(env.MESSAGES_DIR, pkg.ATOM.CPF)
        messages = []
        if os.path.exists(warning_path):
            with open(warning_path, "r") as file:
                messages.append(("WARN", file.read()))

        if os.path.exists(info_path):
            with open(info_path, "r") as file:
                messages.append(("INFO", file.read()))

        if messages:
            print()
            print(">>> " + l10n("pkg-messages", atom=bright(green(pkg))))
            for typ, msg in messages:
                if typ == "WARN":
                    warning(msg)
                elif typ == "INFO":
                    info(msg)
            print()

    for set_name in selected_sets:
        if set_name not in BUILTIN_SETS:
            if deselect or delete or depclean:
                remove_set("selected-sets", Atom(set_name))
            else:
                add_set("selected-sets", Atom(set_name))

    # Commit changes to installed database
    gitrepo = git.Repo.init(env.prefix().INSTALLED_DB)
    try:
        gitrepo.head.commit
    except ValueError:
        gitrepo.git.commit(m=l10n("initial-commit"))

    transstring = io.StringIO()
    print_transactions(merged, verbose=True, out=transstring, summarize=False)
    if gitrepo.git.diff("HEAD", cached=True):
        # There was an error. We report the packages that were successfully merged and
        # note that an error occurred, however we still commit anyway.
        if err:
            gitrepo.git.commit(
                m=(
                    l10n(
                        "merge-success-and-error", num=len(transactions.pkgs), atom=err
                    )
                    + "\n"
                    + transstring.getvalue()
                )
            )
        else:
            gitrepo.git.commit(
                m=(
                    l10n("merge-success", num=len(transactions.pkgs))
                    + "\n"
                    + transstring.getvalue()
                )
            )

    # Check if packages were just modified and can be removed from the rebuild set
    # Any transaction type warrants removal, as they were either rebuilt,
    # and thus can be removed, or deleted, and no longer need to be rebuild
    for atom in get_set("rebuild"):
        installed_pkg = load_installed_pkg(atom)
        if not installed_pkg or installed_pkg.CPN in [
            trans.pkg.CPN for trans in merged.pkgs
        ]:
            remove_set("rebuild", atom)

    info(l10n("checking-rebuild"))
    # Check if packages need to be added to rebuild set
    for pkg in load_all_installed():
        if pkg.CPN not in get_set("rebuild") and pkg.INSTALLED_REBUILD_FILES:
            seen = set()
            for entry in get_rebuild_manifest(
                pkg.get_installed_env().get("REBUILD_FILES", [])
            ):
                seen.add(entry.name)
                if pkg.INSTALLED_REBUILD_FILES.get(entry.name) != entry:
                    add_set("rebuild", pkg.CPN)
                    break
            if not all(entry in seen for entry in pkg.INSTALLED_REBUILD_FILES.entries):
                add_set("rebuild", pkg.CPN)

    if get_set("rebuild"):
        warning(l10n("rebuild-message"))
        for atom in get_set("rebuild"):
            print("    {}".format(green(atom)))
        print(
            l10n(
                "rebuild-prompt",
                command=lgreen(f"portmod {env.PREFIX_NAME} merge @rebuild"),
            )
        )

    global_updates()



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Portmod config tests
"""

import os
import sys
from logging import error

import pytest

from portmod._cli.inquisitor import main, scan_file, scan_root

from .env import TEST_REPO, setup_env, tear_down_env


@pytest.fixture(scope="module", autouse=True)
def setup():
    """
    Sets up and tears down the test environment
    """
    dictionary = setup_env("test")
    yield dictionary
    tear_down_env()


def test_inquisitor():
    """
    Basic inquisitor test on test repo
    """
    has_error = False

    def err(string: str):
        nonlocal has_error
        error(string)
        has_error = True

    scan_root(TEST_REPO.location, err)
    for path_root, _, filenames in os.walk(TEST_REPO.location):
        for filename in filenames:
            scan_file(os.path.join(path_root, filename), TEST_REPO.location, err)
    if has_error:
        raise Exception("Inqusitor failed. See error log for details.")


def test_inquisitor_main():
    """
    Basic inquisitor test on test repo
    """
    sys.argv = ["inquisitor", "scan", TEST_REPO.location]
    main()


# TODO: Add tests that modify the test repo to be invalid



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

# pylint: disable=no-member

"""
Helper functions for interacting with the Windows registry
"""

import os
import sys
from typing import Any, Dict, Optional, Union


def read_reg(
    key: str, subkey: str, entry: Optional[str] = None
) -> Union[Any, Dict[str, Any]]:
    """
    Reads the given registry key/subkey

    args:
        key: Registry key to read from
        subkey: Registry subkey to read from
        entry: Optional name in the dictionary stored at the given key/subkey
               to use.
    returns:
        Key data, type is usually a string. If the key contains subkeys and the entry
        is not provided, returns a dictionary mapping subkey names to their values
    """
    if sys.platform == "win32":
        import winreg  # pylint: disable=import-error

        with winreg.ConnectRegistry(None, key) as reg:
            try:
                rawkey = winreg.OpenKey(
                    reg, subkey, access=winreg.KEY_READ | winreg.KEY_WOW64_64KEY
                )
            except FileNotFoundError:
                try:
                    rawkey = winreg.OpenKey(
                        reg, subkey, access=winreg.KEY_READ | winreg.KEY_WOW64_32KEY
                    )
                except FileNotFoundError:
                    return None

            if entry is None:
                subkeys = {}
                i = 0
                try:
                    while True:
                        subsubkey = winreg.EnumKey(rawkey, i)
                        subkeys[subsubkey] = read_reg(
                            key, subkey + os.sep + subsubkey, entry
                        )
                        i += 1
                except WindowsError:  # pylint: disable=undefined-variable
                    if subkeys:
                        return subkeys
            try:
                i = 0
                while True:
                    name, value, _ = winreg.EnumValue(rawkey, i)
                    if entry is None:
                        return value
                    if name == entry:
                        return value
                    i += 1
            except WindowsError:  # pylint: disable=undefined-variable
                return None
    else:
        raise Exception("read_reg should not be called on platforms other than win32!")



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""
Module that provides a function to aid construction of safe
wrapper libraries around unsafe system libraries
"""
from types import SimpleNamespace
from typing import List, Optional


class Permissions(SimpleNamespace):
    def __init__(
        self,
        copy: Optional["Permissions"] = None,
        *,
        rw_paths: List[str] = [],
        ro_paths: List[str] = [],
        global_read: Optional[bool] = None,
        network: Optional[bool] = None,
        tmp: Optional[str] = None,
    ):
        if copy:
            self.rw_paths: List[str] = copy.rw_paths
            self.ro_paths: List[str] = copy.ro_paths
            self.global_read: Optional[bool] = copy.global_read
            self.network: Optional[bool] = copy.network
            self.tmp: Optional[str] = copy.tmp
        else:
            self.global_read = False
            self.network = False
            self.rw_paths = []
            self.ro_paths = []
            self.tmp = None

        self.rw_paths.extend(rw_paths)
        self.ro_paths.extend(ro_paths)
        if global_read is not None:
            self.global_read = global_read
        if network is not None:
            self.network = network
        if tmp is not None:
            self.tmp = tmp



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

import enum
import os
from typing import Any, Dict

from portmod.globals import env
from portmod.l10n import l10n


def get_archive_basename(archive: str) -> str:
    """Returns archive name minus extension(s)"""
    basename, _ = os.path.splitext(archive)
    # Hacky way to handle tar.etc having multiple extensions
    if basename.endswith("tar"):
        basename, _ = os.path.splitext(basename)
    return basename


class EnumMeta(enum.EnumMeta):
    def __contains__(cls, item: Any):
        return item in {member.value for member in cls.__members__.values()}  # type: ignore


class HashAlg(enum.Enum, metaclass=EnumMeta):
    """
    Class for interacting with supported hash algorithms that can be used in manifests
    """

    BLAKE2B = "BLAKE2B"
    MD5 = "MD5"
    SHA512 = "SHA512"
    BLAKE3 = "BLAKE3"

    def __lt__(self, other):
        return self.value < other.value


class LocalHashError(Exception):
    """Exception indicating an unexpected download file"""


class Source:
    """
    Class used for storing information about download files without manifest information
    """

    def __init__(self, url: str, name: str):
        self.url = url
        self.name = name
        self.path = os.path.join(env.DOWNLOAD_DIR, name)
        self.basename = get_archive_basename(name)

    def __repr__(self):
        return self.url

    def to_json(self):
        return {"url": self.url, "name": self.name}

    def __eq__(self, other):
        if not isinstance(other, Source):
            return False
        return self.url == other.url and self.name == other.name

    def __hash__(self):
        return hash((self.url, self.name))


class SourceManifest(Source):
    """Class used for storing information about download files"""

    def __init__(self, source: Source, hashes: Dict[HashAlg, str], size: int):
        super().__init__(source.url, source.name)
        self.hashes = hashes
        self.size = size

    def __hash__(self):
        return hash((self.url, self.name, tuple(self.hashes)))

    def as_source(self):
        return Source(self.url, self.name)

    def check_file(self, filename: str, raise_ex=False) -> bool:
        """
        Returns true if and only if the hash of the given file
        matches the stored hash
        """
        from portmod.fs.util import get_hash

        hashes_to_check = sorted(self.hashes)
        if len(hashes_to_check) > 1 and HashAlg.MD5 in hashes_to_check:
            # Ignore MD5 unless it's the only hash. It's neither particularly fast, nor reliable
            # It's only used since certain services will supply an MD5 hash which we can compare to
            hashes_to_check.remove(HashAlg.MD5)
        results = get_hash(filename, tuple(sorted(self.hashes)))
        for halg, result in zip(sorted(self.hashes), results):
            if self.hashes[halg] != result:
                if raise_ex:
                    raise LocalHashError(
                        l10n(
                            "local-hash-mismatch",
                            filename=filename,
                            hash=halg.name,
                            hash1=self.hashes[halg],
                            hash2=result,
                        )
                    )
                return False
        return True



# Copyright 2019-2020 Portmod Authors
# Distributed under the terms of the GNU General Public License v3

"""Module for reading from list files"""

import os
from typing import List, Optional


def read_list(listpath: str, encoding: Optional[str] = None) -> List[str]:
    """Reads the given list file and returns its contents"""
    with open(listpath, mode="r", encoding=encoding) as list_file:
        return [line.strip() for line in list_file.read().splitlines() if line]


def add_list(listpath: str, entry: str):
    """Appends the given value to the list file"""
    os.makedirs(os.path.dirname(listpath), exist_ok=True)
    with open(listpath, mode="a") as list_file:
        print(entry, file=list_file)


def write_list(path: str, contents: List[str]):
    """Writes a list file containing the given list"""
    with open(path, "w") as file:
        for value in contents:
            print(value, file=file)
